(* This library defines an executable version of the SCC algorithm. *)

Set Implicit Arguments.
Require Import List.
Require Import LibTactics.
Require Import MyTactics.
Require Import MySets.
Require Import MyRelations.
Require Import DFS.
Require Import SigSET.
Require Import SigFOREACH.
Require Import DFSExec.
Require Import SCC.

(* ---------------------------------------------------------------------------- *)

Section SCCExec.

  Variable V : Type.
  Variable E : V -> V -> Prop.

  Variable finiteness:
    finite (@universe V).

  Variable VSet : SET V.

  Variable successor : V -> FOREACH V.
  Variable successor_spec :
    forall v,
    FOREACH_SET_SPEC (successor v) (image E (singleton v)).

  Variable predecessor : V -> FOREACH V.
  Variable predecessor_spec :
    forall v,
    FOREACH_SET_SPEC (predecessor v) (image (reverse E) (singleton v)).

  Variable vertex : FOREACH V.
  Variable vertex_spec :
    FOREACH_SET_SPEC vertex (@universe V).

(* ---------------------------------------------------------------------------- *)

  Definition traverse :=
    fold_visitf_initial_implicit finiteness VSet.

  Definition scc_forest : forest V :=
    let (_, forest1) := traverse successor vertex in
    let vertex := MkFOREACH (fun A (accu : A) (f : A -> V -> A) =>
      fold_preorder f forest1 accu
    )
    in
    let (_, forest2) := traverse predecessor vertex in
    forest2.

End SCCExec.

(* Optional.
Extraction traverse.
Extraction scc_forest. (* TEMPORARY possible extraction bug? *)
*)

