(* This file defines a type [SET V] of machine representations of sets of
   elements of type [V]. *)

Set Implicit Arguments.
Require Import MySets.

Record SET (V : Type) := MkSET {

  (* A set is represented at runtime as a value of type [repr]. The function
     [meaning] maps such a value to its interpretation as a mathematical set. *)

  repr : Type;
  meaning : repr -> set V;

  (* The constant [void] represents the empty set. *)

  void : repr;
  void_spec : meaning void = empty_;

  (* The function [mark] inserts an element into a set. *)

  mark : V -> repr -> repr;
  mark_spec :
    forall v marks,
    meaning (mark v marks) = insert v (meaning marks);

  (* The function [marked] tests membership in a set. *)

  marked : V -> repr -> bool;
  marked_spec:
    forall v marks,
    marked v marks = true <-> member v (meaning marks)

}.
